const isQuoted = (s) => {
  const first = s[0];
  const last = s[s.length - 1];
  return s.length > 1 && first === last && (first === '"' || last === "'");
};
class Iterator {
  constructor(input) {
    this.input = input;
    this.start = 0;
    this.current = 0;
  }
  error(err) {
    throw new Error(`Fenceparser: ${err}.`);
  }
  peek(n = 0) {
    return this.input[this.current + n];
  }
  advance() {
    return this.input[this.current++];
  }
  isAtEnd() {
    return this.current >= this.input.length;
  }
}

const lex = (input) => new Lexer(input).scan();
const KEYWORDS = {
  true: true,
  false: false
};
const isAlpha = (s) => /[a-zA-Z_$-]/.test(s);
const isNumeric = (s) => /[0-9]/.test(s);
const isAlphaNumeric = (s) => isAlpha(s) || isNumeric(s);
class Lexer extends Iterator {
  constructor() {
    super(...arguments);
    this.output = [];
  }
  string(quote) {
    while (this.peek() !== quote && !this.isAtEnd())
      this.advance();
    if (this.isAtEnd()) {
      this.error("Unterminated string");
    }
    this.advance();
    this.output.push(this.input.substring(this.start, this.current));
  }
  number() {
    while (isNumeric(this.peek()))
      this.advance();
    if (this.peek() === "-" && isNumeric(this.peek(1))) {
      this.advance();
      while (isNumeric(this.peek()))
        this.advance();
      this.output.push(this.input.substring(this.start, this.current));
      return;
    } else if (this.peek() === "." && isNumeric(this.peek(1))) {
      this.advance();
      while (isNumeric(this.peek()))
        this.advance();
    }
    this.output.push(parseFloat(this.input.substring(this.start, this.current)));
  }
  identifier() {
    while (isAlphaNumeric(this.peek()) && !this.isAtEnd())
      this.advance();
    const text = this.input.substring(this.start, this.current);
    if (Object.keys(KEYWORDS).includes(text)) {
      this.output.push(KEYWORDS[text]);
    } else {
      this.output.push(text);
    }
  }
  scan() {
    while (!this.isAtEnd()) {
      this.start = this.current;
      const next = this.advance();
      switch (next) {
        case "{":
        case "}":
        case "=":
        case ",":
        case ":":
        case "[":
        case "]":
          this.output.push(next);
          break;
        case '"':
        case "'":
          this.string(next);
          break;
        case " ":
        case "\r":
        case "	":
        case "\n":
          break;
        default:
          if (isNumeric(next)) {
            this.number();
          } else if (isAlpha(next)) {
            this.identifier();
          } else {
            this.error(`Unexpected character ${next}`);
          }
      }
    }
    return this.output;
  }
}

const parse = (input) => new Parser(input).parse();
class Parser extends Iterator {
  constructor() {
    super(...arguments);
    this.output = {};
  }
  object() {
    const result = {};
    const parseValue = () => {
      let identifier = this.advance();
      if (typeof identifier === "string" && isQuoted(identifier)) {
        identifier = identifier.slice(1, -1);
      } else {
        identifier = identifier.toString();
      }
      if (this.peek() === ":") {
        this.advance();
        result[identifier] = this.value();
      } else {
        result[identifier] = true;
      }
    };
    this.advance();
    if (this.peek() !== "}") {
      parseValue();
      while (this.peek() === ",") {
        this.advance();
        if (this.peek() === "}") {
          this.error("Trailing comma");
        }
        parseValue();
      }
    }
    if (this.advance() !== "}") {
      this.error("Unterminated object");
    }
    return result;
  }
  array() {
    const result = [];
    this.advance();
    if (this.peek() !== "]") {
      result.push(this.value());
      while (this.peek() === ",") {
        this.advance();
        if (this.peek() === "]") {
          this.error("Trailing comma");
        }
        result.push(this.value());
      }
    }
    if (this.advance() !== "]") {
      this.error("Unterminated array");
    }
    return result;
  }
  value() {
    if (this.peek() === "{") {
      return this.object();
    } else if (this.peek() === "[") {
      return this.array();
    } else if (typeof this.peek() === "string" && isQuoted(this.peek())) {
      return this.advance().slice(1, -1);
    } else {
      return this.advance();
    }
  }
  parse() {
    if (this.input.length < 1) {
      return null;
    }
    while (!this.isAtEnd()) {
      const peeked = this.peek();
      if (peeked === "{") {
        if (!("highlight" in this.output)) {
          this.output["highlight"] = {};
        }
        this.output["highlight"] = {
          ...this.output["highlight"],
          ...this.object()
        };
      } else {
        const identifier = this.advance();
        if (this.peek() === "=") {
          this.advance();
          this.output[identifier] = this.value();
        } else {
          this.output[identifier] = true;
        }
      }
    }
    return this.output;
  }
}

const index = (input) => parse(lex(input));

export { index as default, lex, parse };
